#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include "lkh.h"
#include "tree.h"
#include "key.h"
#include "lkhutil.h"
#include "classiclkh.h"


int ClassicLeave(alberolkh *radice, Sij darevocare);
int ClassicJoin(alberolkh *radice, Sij dainserire, int maxfigli);
void EliminaPath(alberolkh *apartireda);
long NumFoglie(alberolkh *radice);
long ClassicRekeying(alberolkh *radice, nodosd *utenti, long nutenti);
alberolkh *AlberoClassicLkh(long utenti, long figli);
int CreaPath(alberolkh **nodo, alberolkh *padre, long altezzaattuale, int figli);
alberolkh *CreaAlberoCompleto(long altezza, long nfigli);
int ClassicBroadcast(chiave nuova, long *destinatari, long ndest, nodosd *utenti, long nutenti);
long Foglie(alberolkh *radice, long **foglie);
void Ripristina(alberolkh *radice);
void AbilitaUtenteClassicLkh(nodosd *utenticlkh, long nutenti, long da_abilitare);


/*
*
*
	FUNZIONI DI INTERFACCIA DEL CLASSIC LKH
*
*
*/

/* InitClassicLkh(): inizializza le strutture dati necessarie al funzionamento dell'algoritmo LKH classico (quindi solamente l'albero LKH, questa volta statico al contrario dell'LKH utilizzato nell'algoritmo SD+LKH). */
int InitClassicLkh(lkh *centralelkh, long primoutente, long nutenti, int maxfigli, int sbil, nodosd *utenticlkh)
{
	alberolkh *radice;
	chiave iniziale="algoritmoLKH";
	Sij utente;
	long i;


        centralelkh->d=maxfigli; /* fisso il numero di figli max per nodo*/
	centralelkh->c=sbil;
	centralelkh->subsets=NULL; /* i subsets non sono usati dall'LKH classico */

	/* alloca la radice dell'albero dell'lkh controlla che l'allocazione vada a buon fine*/
	if((radice=malloc(sizeof(alberolkh)))==NULL)
	{
		return FALSE;
	}

	/* inizializza opportunamente i campi della radice, se ci sono ERROREi di allocazione dealloca tutto cio' che e' stato precedentemente allocato ed esce restituendo al chiamante FALSE */
	if((radice->sons=calloc(maxfigli,sizeof(alberolkh *)))==NULL)
	{
		free(radice);
		return FALSE;
	}
	radice->nfigli=0; /* la radice per ora non ha figli*/
	radice->mysubset=NULL; /* la radice e' un nodo interno, dunque non ha subset */
	/* la chiave non viene inizializzata poiche' al primo inserimento nell'albero la chiave della radice dovra' essere riassegnata*/
	radice->iscorrupted=FALSE; /* il path cui appartiene la radice inizialmente non e' corrotto */
	radice->parent=NULL; /* la radice in quanto tale non ha alcun genitore */

	memcpy(radice->key,iniziale,SHA_DIGEST_LENGTH);
	centralelkh->root=radice;


	/* Inserisce gli utenti (per compatibilita' con la funzione LKH utilizzata nell'algoritmo SD+LKH si adopera la seguente convenzione: l'user x-esimo e' indicato con il subset avente i=x e j=x) nell'albero LKH) */

	for(i=primoutente;i<(primoutente+nutenti);i++)
	{
		/* fa' si' che <utente> rappresenti l'utente i-esimo secondo la convenzione specificata in precedenza */
		utente.i=utente.j=i;
		//printf("inserisco l'utente %ld\n",i);

		/*inserisce l'utente i-esimo all'interno dell'albero LKH */
		if(Inserisci(centralelkh->root,utente,maxfigli)==FALSE)
		{
			printf("Errore nell'inserimento di %ld\n",utente.i);
			return ERRORE;
		}
		Ripristina(centralelkh->root); /* l'albero ha tutti i path corrotti perche' e' stato appena creato, quindi viene ripristinato uno stato consistente con la simulazione che sara' avviata dopo init */

	}
	
	/* se il numero di utenti non e' una potenza di due, allora revoca direttamente gli utenti in eccesso (ovvero quelli che gia' prima del round 1 non sono considerati abilitati) senza inserirli nell'albero LKH. Questi utenti potranno essere abilitati (ed entrare nell'albero LKH) se e' attiva la funzione di inserimento utenti. */
	for(i=(primoutente+nutenti);i<primoutente*2+1;i++)
	{
		//printf("Revoco all'inizio %ld che diventa %ld\n",i,i-primoutente+2);
		/* dall'etichetta del primoutente e dall'etichetta dell'utente da revocare, ottiene l'indice dell'utente nell'array degli utenti lkh*/
		utenticlkh[i-primoutente].revocato=TRUE;
		utenticlkh[i-primoutente].shortkeys=NULL;
	}


	//Stampa(centralelkh->root);

	return TRUE;
}

/* ClassicLkh(): implementa il funzionamento dell'algoritmo classico LKH cosi' come e' stato descritto dai suoi inventori.*/
long ClassicLkh(lkh centrale, l_revocati *revocati, enviroment ambiente, nodosd *utenti, long nutenti)
{
	long i;
	long nmessaggi;
	long nrevocati;
	sij revocato;
	FILE *bilancia;
	
	/* il numero dei revocati e' (in teoria) pari al lotto da revocare dell'lkh, ma nell'ultima iterazione cio' potrebbe non essere vero, poiche' potrebbero essere rimasti meno di <lotto_lkh> utenti da revocare: dunque per sicurezza si prende come numero di revocati il minimo tra i revocati effettivi e il lotto scelto. */
	nrevocati=Min((revocati->tot_revocati-revocati->primo_lkh),ambiente.lotto_lkh);
	
	nmessaggi=0; /* il numero di messaggi spediti per ora e' zero*/

	/* per ogni user da revocare... */
	for(i=revocati->primo_lkh;i<(revocati->primo_lkh+nrevocati);i++)
	{
		/*...revoca effettivamente l'user */
		//printf("revoco l'utente %ld, i=%ld, nutenti=%ld\n",(revocati->revocati)[i],i,nutenti);
		RevocaClassicLkh(utenti,nutenti,(revocati->revocati)[i]);
		
		/* elimina l'user dall'albero LKH e ne corrompe il path */
		revocato.i=(revocati->revocati)[i];
		revocato.j=(revocati->revocati)[i];
		
		if(Delete(centrale.root,revocato)==FALSE)
		{
			printf("ERRORE su ClassicLeave cercavo il nodo %ld alla revoca %ld-esima\n",revocato.i,i);
			return ERRORE;
		}
	}
	
	revocati->primo_lkh+=nrevocati; /* il numero totale di revocati e' ora pari a primo_lkh+nrevocati (in virtu' del ciclo precedente)*/
	
	/* prima di bilanciare simula l'esecuzione del Rekeying, in modo da separare i messaggi che dipendono dal bilanciamento da quelli che dipendono dall'algoritmo vero e proprio */
	
	if((centrale.root)->nfigli!=0)
	{
		nmessaggi=SimulaRekeying(centrale.root, utenti, nutenti,0);
  	}
	
		

	/* a seguito degli inserimenti o delle cancellazioni effettuate, l'albero potrebbe essere sbilanciato, chiama dunque Balance() che provvedera' a controllare il bilanciamento, e nel caso a bilanciare l'albero */

	i=Balance(centrale.root,centrale.c,centrale.d);
	if(i==FALSE)
	{
		return ERRORE;
	}
	/*
	else if(i==2)
	{
		printf("B");
	}*/

	/* riassegna le chiavi dei path corrotti secondo la filosofia LKH, a patto che nell'albero sia presente almeno un utente */
	if((centrale.root)->nfigli!=0)
	{
		i=Rekeying(centrale.root, utenti, nutenti,0);
  	}
	

	if(nmessaggi==ERRORE || i==ERRORE)
	{
		printf("ERRORE su ClassicRekeying\n");
		return ERRORE;
	}
	
	/* salva su file i messaggi che dipendono dal bilanciamento */
	if((bilancia=fopen("bil_lkh.dat","a"))==NULL)
	{
		printf("ERRORE impossibile salvare i dati relativi al bilanciamento\n");
		return ERRORE;
	}
	
	fprintf(bilancia,"%ld\n",i); /* il numero di mex dovuti al bilanciamento e' pari al numero di mex totali meno quelli dovuti esclusivamente all'algoritmo senza bilanciamento.*/
	fclose(bilancia);

	return nmessaggi;
}

int AbilitaClassicLkh(lkh centralelkh,nodosd *utenti,long nutenti,long prossimo,long abilitati)
{
	Sij utente;
	long i;

	/* per ognuno degli utenti da abilitare...*/
	for(i=prossimo;i<(prossimo+abilitati);i++)
	{
		/*...abilita fisicamente l'utente */
		AbilitaUtenteClassicLkh(utenti,nutenti,i);
		
		/* fa' si' che <utente> rappresenti l'utente i-esimo secondo la convenzione specificata in precedenza */
		utente.i=utente.j=i+nutenti-1;
		
		//printf("abilito %ld che diventa %ld\n",i,i+nutenti-1);
		
		/*inserisce l'utente i-esimo all'interno dell'albero LKH */
		if(Inserisci(centralelkh.root,utente,centralelkh.d)==FALSE)
		{
			printf("Errore nell'inserimento di %ld\n",utente.i);
			return ERRORE;
		}
		
	}
	
	return TRUE;
}

/*
*
*
	FUNZIONI PER LA CREAZIONE DELL'ALBERO DEL CLASSIC LKH
*
*
*/


/*AlberoClassicLkh(): crea l'albero del classic LKH; un albero completo di altezza t.c. e' possibile inserire sullo stesso livello un numero di foglie pari ad <utenti> in modo che l'albero rimanga completo tranne al piu' l'ultimo livello. */
alberolkh *AlberoClassicLkh(long utenti, long figli)
{
	long altezza;
	long u,f;
	double logaritmo_u_inbase_f;

	u=utenti;
	f=figli;

	/* per calcolare log(u) in base f applica la regola del cambiamento di base */
	logaritmo_u_inbase_f=log(u)/log(f);

	/* calcola l'altezza necessaria perche' l'albero n-ario (con n=figli) sia un albero completo a meno (eventualmente) dell'ultimo livello (quello che conterra' le foglie, ovvero gli utenti). In pratica, calcola il logaritmo di <utenti> in base <figli>, arrotondandolo al piu' piccolo intero non minore del logaritmo. */
	altezza=(long) ceil(logaritmo_u_inbase_f);


	/* ritorna al chiamante un albero completo di (altezza-1): l'ultimo livello (quello di profondita' <altezza>) sara' quello che ospitera' le foglie (il cui inserimento e' demandato al chiamante) */
	return CreaAlberoCompleto(altezza-1,figli);
}

/* CreaAlberoCompleto(): crea un albero n-ario (con n=nfigli) completo di profondita' <altezza>*/
alberolkh *CreaAlberoCompleto(long altezza, long nfigli)
{
	alberolkh *radice;
	int i;

	/* alloca la radice dell'albero completo */

	if((radice=malloc(sizeof(alberolkh)))==NULL)
	{
		return NULL;
	}

	if((radice->sons=calloc(nfigli,sizeof(alberolkh *)))==NULL)
	{
		free(radice);
		return NULL;
	}

	radice->nfigli=0; /* la radice per ora non ha figli*/
	radice->mysubset=NULL; /* la radice e' un nodo interno, dunque non ha subset */
	/* la chiave non viene inizializzata poiche' al primo inserimento nell'albero la chiave della radice dovra' essere riassegnata*/
	radice->iscorrupted=FALSE; /* il path cui appartiene la radice inizialmente non e' corrotto */
	radice->parent=NULL; /* la radice in quanto tale non ha alcun genitore */

	if(altezza==1) /* se l'albero dev'essere di altezza 1 allora la creazione e' completa*/
	{
		return radice;
	}

	/* altrimenti, per ogni figlio di radice, si crea un path completo di profondita' pari a (altezza-1), essendo gia' stato creato, con la radice, il primo livello */
	for(i=0;i<nfigli;i++)
	{
		if(CreaPath(&((radice->sons)[i]),radice,altezza-1,nfigli)==ERRORE)
		{
			return NULL;
		}
	}

	return radice;

}

/* CreaPath(): crea un path completo di altezza pari ad <altezzaattuale> di un albero n-ario (con n=nfigli) */
int CreaPath(alberolkh **nodo, alberolkh *padre, long altezzaattuale, int figli)
{
	int i;

	/* alloca il nuovo nodo */

	if((*nodo=malloc(sizeof(alberolkh)))==NULL)
	{
		return ERRORE;
	}

	if(((*nodo)->sons=calloc(figli,sizeof(alberolkh *)))==NULL)
	{
		free(nodo);
		return ERRORE;
	}

	(*nodo)->nfigli=0; /* la (*nodo) per ora non ha figli*/
	(*nodo)->mysubset=NULL; /* la (*nodo) e' un nodo interno, dunque non ha subset */
	/* la chiave non viene inizializzata poiche' al primo inserimento nell'albero la chiave della (*nodo) dovra' essere riassegnata*/
	(*nodo)->iscorrupted=FALSE; /* il path cui appartiene la (*nodo) inizialmente non e' corrotto */
	(*nodo)->parent=padre; /* collega il nodo al padre */

	padre->sons[padre->nfigli]=*nodo; /* collega il padre al figlio */
	(padre->nfigli)++; /* incrementa il numero di figli del padre */


	if(altezzaattuale==0) /* se il path e' stato creato per la profondita' richiesta...*/
	{
		return TRUE; /* termina la funzione restituendo TRUE, poiche' tutto e' andato a buon fine */
	}

	/* altrimenti, per ogni figlio di nodo, si crea un path completo di profondita' pari all'altezza rimanente (altezzaattuale-1) */
	for(i=0;i<figli;i++)
	{
		if(CreaPath(&(((*nodo)->sons)[i]),*nodo,altezzaattuale-1,figli)==ERRORE)
		{
			return ERRORE;
		}
	}

	return TRUE;

}

/*
*
*
	FUNZIONI DI GESTIONE DELL'ALBERO DEL CLASSIC LKH
*
*
*/

/*ClassicJoin(): inserisce nell'albero radicato in <radice> un nodo con Sij <dainserire>. Se l'albero in cui si inserisce e' completo tranne al piu' l'ultimo livello (ed esiste, almeno nell'ultimo livello, un posto libero ove inserire il nuovo nodo), allora tale proprieta' viene mantenuta anche dopo l'inserimento. */
int ClassicJoin(alberolkh *radice, Sij dainserire, int maxfigli)
{
	alberolkh *padre,*figlio;
	int a;

	/* cerca il padre del nuovo nodo da creare. Il padre scelto sara' il nodo di altezza minore e piu' "a sinistra" dell'albero tra quelli con un numero di figli minore di <maxfigli> (dunque nodi in cui e' possibile ancora inserire un figlio) */
	if((padre= LastLeaf(radice,&a,maxfigli))==NULL)
	{
		return ERRORE;
	}

	/* alloca il nuovo nodo */

	if((figlio=malloc(sizeof(alberolkh)))==NULL)
	{
		return ERRORE;
	}

	if((figlio->sons=calloc(maxfigli,sizeof(alberolkh *)))==NULL)
	{
		free(radice);
		return ERRORE;
	}

	if((figlio->mysubset=malloc(sizeof(Sij)))==NULL)
	{
		return ERRORE;
	}





	figlio->nfigli=0; /* la figlio per ora non ha figli*/
	(figlio->mysubset)->i=dainserire.i;
	(figlio->mysubset)->j=dainserire.j;
	/* la chiave non viene inizializzata poiche' al primo inserimento nell'albero la chiave della figlio dovra' essere riassegnata*/
	figlio->iscorrupted=FALSE; /* il path cui appartiene la figlio inizialmente non e' corrotto */
	figlio->parent=padre; /* collega il nodo al padre */

	padre->sons[padre->nfigli]=figlio; /* collega il padre al figlio */
	(padre->nfigli)++; /* incrementa il numero di figli del padre*/

	return TRUE;
}



/* ClassicLeave(): revoca l'user con l'Sij darevocare e ne corrompe il path, eliminandolo del tutto (a meno della radice) nel caso in cui non abbia al suo interno alcun user. */
int ClassicLeave(alberolkh *radice, Sij darevocare)
{
	int i;
	alberolkh *padre;


	/* se la radice e' NULL...*/
	if(radice==NULL)
	{
		/*..ritorna FALSE, poiche' non c'e' nodo da cancellare*/
		return FALSE;
	}
	/*altrimenti, se <radice> e' una foglia (con subset)...*/
	else if(IsLeaf(radice) && radice->mysubset!=NULL)
	{
		/*... se il subset di radice e' proprio quello che va cancellato...*/
		if((radice->mysubset)->i==darevocare.i && (radice->mysubset)->j==darevocare.j)
		{
			padre=radice->parent;
			/*...cerca il puntatore che lega il padre alla foglia da cancellare */
			for(i=0;i<padre->nfigli && (padre->sons)[i]!=radice;i++);
			/* alla fine del ciclo, in (padre->sons)[i] c'e' il puntatore al nodo da cancellare*/
			(padre->sons)[i]=(padre->sons)[padre->nfigli-1]; /* sovrascrive il puntatore al nodo da cancellare con il puntatore all'ultimo dei figli di padre */

			/* dealloca il nodo radice */
			free(radice->sons);
			free(radice->mysubset);
			free(radice);

			/* decrementa di una unita' il numero di figli di <padre>*/
			(padre->nfigli)--;

			if(padre->nfigli==0)
			{
				EliminaPath(padre);
			}
			else
			{/* il path di radice viene corrotto */
				Corrompi(padre);
			}

			return TRUE;
		}
		/* altrimenti, se la foglia in questione NON e' l'insieme Sij desiderato...*/
		else
		{
			/*...restituisce FALSE, poiche' in questo sottoalbero non e' presente tale insieme*/
			return FALSE;
		}
	}
	/* se invece <radice> e' un nodo interno, allora bisogna scandire tutti i sottoalberi radicati in ognuno dei suoi figli, per vedere se uno di essi contiene l'insieme Sij cercato*/
	else
	{
		/*per ogni figlio di <radice>...*/
		for(i=0;i<radice->nfigli;i++)
		{
			/*...se nel sottoalbero radicato nell'i-esimo figlio di <radice> la cancellazione ha esito positivo...*/

			if(ClassicLeave((radice->sons)[i],darevocare)==TRUE)
			{
				/*...restituisce TRUE per segnalare l'avvenuta cancellazione*/
				return TRUE;
			}
		}
		/* se il ciclo e' terminato senza aver trovato l'Sij da cancellare, restituisce FALSE per segnalare che l'Sij in questione non e' presente nel sottoalbero corrente */
		return FALSE;
	}

}


long ClassicRekeying(alberolkh *radice, nodosd *utenti, long nutenti)
{
	int i;
	alberolkh *figlio;
	long nmex,nmexfiglio;
	long *dest,*tmp,quanti;




	/* se la radice non esiste, oppure non e' corrotta...*/
	if(radice == NULL || !(radice->iscorrupted))
	{
		/* ... allora non occorre inviare messaggi, dunque il numero di mex e' 0*/
		return 0;
	}

	/* se la radice e' una foglia...*/
	if(IsLeaf(radice))
	{
		/*...allora, poiche' il genitore sta per subire il rekeying, la foglia non va piu' considerata corrotta */
		//if(radice->mysubset!=NULL) printf("Passo per %ld\n",radice->mysubset->i);
		radice->iscorrupted=FALSE;
		return 0;
	}


	/* assegna a <radice> una nuova chiave*/
	if ( !GetNewKey(radice->key) )
	{
		printf("In ClassicREkeying ERRORE su GetNewKey\n");
		return ERRORE;
	}
	radice->iscorrupted=FALSE;

	nmex=0;

	/*per ogni figlio di <radice>...*/
	for(i=0;i<radice->nfigli;i++)
	{
		figlio=(radice->sons)[i];
		/*  se il figlio i-esimo di <radice>  appartiene ad un path corrotto...*/
		if(figlio->iscorrupted==TRUE)
		{
			/*... prima di riassegnare la propria chiave, procede a riassegnare le chiavi del path di <figlio>*/
			nmexfiglio=ClassicRekeying(figlio,utenti,nutenti);
			if(nmexfiglio==ERRORE)
			{
				printf("In ClassicREkeying ERRORE sulla chiamata ricorsiva\n");
				return ERRORE;
			}
			/* il numero di messaggi inviati per eseguire il rekeying del path del figlio va aggiunto al numero di messaggi totali*/
			nmex+=nmexfiglio;
		}

		/* se l'albero radicato nel figlio contiene foglie (con subset) ovvero utenti non revocati...*/
		if(figlio->iscorrupted!=ERRORE)
		{
			 /*invia in Broadcast la nuova chiave di <radice> crittata con la chiave di <figlio>: dunque questo messaggio sara' decifrabile solo dalle foglie del sottoalbero di radice radicato in <figlio>*/
			dest=calloc(nutenti,sizeof(long));
			tmp=dest;
			quanti=Foglie(figlio, &tmp);
			if(ClassicBroadcast(radice->key,dest,quanti,utenti,nutenti)==ERRORE)
			{
				printf("In ClassicREkeying ERRORE su ClassicBroadcast\n");
				return ERRORE;
			}

			/* essendo appena stato spedito un messaggio, incrementa di un'unita' il numero di messaggi spediti*/
			nmex++;

			/* dealloca la memoria utilizzata per <destinatari>*/
			free(dest);

		}

	}

	return nmex;

}

int ClassicBroadcast(chiave nuova, long *destinatari, long ndest, nodosd *utenti, long nutenti)
{
	listachiavi *l;
	long user;
	int i;


	/* finche' esistono utenti cui inviare in broadcast la nuova chiave...>*/
	for(i=0;i<ndest;i++)
	{
		/* assegna all'utente con indice <user> la nuova chiave*/
		user=destinatari[i];
		/* alloca lo spazio per la nuova chiave nella lista delle chiavi a breve termine dell'utente <user> */
  		if(!(l=malloc(sizeof(listachiavi))))
		{
			return ERRORE;
		}

		memcpy(l->valore,nuova,SHA_DIGEST_LENGTH);
		/* inserisce la nuova chiave in testa alla lista delle chiavi a breve termine di <user> */
		l->next=(utenti[user]).shortkeys;
		(utenti[user]).shortkeys=l;

	}

	return TRUE;

}


/*
*
*
	FUNZIONI DI UTILITA' SULL'ALBERO DEL CLASSIC LKH
*
*
*/



void EliminaPath(alberolkh *apartireda)
{
	alberolkh *daeliminare;


	daeliminare=apartireda;

	/* Finche' ci sono nodi x t.c. il sottoalbero radicato in x non ha foglie (ovvero nodi foglie con subset non nullo), e finche' non si e' giunti alla radice...*/
	while( (daeliminare->parent!=NULL) && NumFoglie(daeliminare)==0 )
	{
		daeliminare->iscorrupted=ERRORE;
		daeliminare=daeliminare->parent;
	}

	Corrompi(daeliminare); /* il path superiore e' corrotto, poiche' e' necessario cambiare le chiavi per evitare che l'utente appena revocato possa accedere alle nuove comunicazioni*/
}



/*Foglie(): pone nell'array la cui prima cella e' puntata da *foglie le foglie dell'albero radicato in <radice>; torna il numero delle stesse.*/
long Foglie(alberolkh *radice, long **foglie)
{
	int i;
	long totale;

	totale=0;


	/* se la radice non esiste...*/
	if(radice == NULL)
	{
		/* ... allora non ci sono foglie nel sottoalbero */
		return 0;
	}
	/* se la radice e' una foglia (con subset)...*/
	if(IsLeaf(radice))
	{
		if( (radice->mysubset)!=NULL )
		{
			/*... inserisce la foglia nella cella corrente dell'array */
			**foglie=radice->mysubset->i;
			/* sposta il puntatore alla cella corrente dell'array in modo che il prossimo inserimento avvenga nella cella successiva */
			(*foglie)++;
			/* il numero totale di foglie e' pari ad uno */
			return 1;
		}
		/* se e' una foglia senza subset, allora e' un nodo interno divenuto foglia, e dunque non va annoverato nel numero totale */
		else
		{
			return 0;
		}
	}

	totale=0;

	/* altrimenti, per ogni figlio di <radice>...*/
	for(i=0;i<radice->nfigli;i++)
	{
		/*...calcola il numero di foglie contenute nel sottoalbero radicato nell'i-esimo figlio */

		totale+=Foglie((radice->sons)[i],foglie);
	}

	return totale;
}



void Ripristina(alberolkh *radice)
{
	int i;

	/* se la radice esiste...*/
	if(radice != NULL)
	{

		/*...se e' corrotta...*/
		if(radice->iscorrupted==TRUE)
		{
			/* allora la rende non corrotta */
			radice->iscorrupted=FALSE;
		}


		/* se la radice non e' una foglia...*/
		if(!IsLeaf(radice))
		{
			/* per ogni figlio di <radice>...*/
			for(i=0;i<radice->nfigli;i++)
			{
				/*... applica la funzione Ripristina() su di esso*/
				Ripristina((radice->sons)[i]);
			}
		}

	}
}

/*
*
*
	FUNZIONI PER LA REVOCA DEGLI UTENTI DEL CLASSIC LKH E PER LA VERIFICA DEL CORRETTO FUNZIONAMENTO DELL'ALGORITMO
*
*
*/


/*RevocaClassicLkh(): revoca <da_rev> utenti del classic LKH e pone i relativi indici in <revocati>*/
/*long RevocaClassicLkh(nodosd *utenti, long da_rev, long att_rev, long nut, long *rev,Sij *rev_r)
{
     long i;
     long j=0;



     for(i=att_rev;i<(att_rev+da_rev) && i<nut;i++)
     {
     	//printf("\n rrev HSD : %ld, rev LKH: %ld\n",rev[i],rev[i]-(nut-1));
     	rev_r[j].i=rev_r[j].j=rev[i]-(nut-1);
	utenti[rev[i]-(nut-1)].revocato=TRUE;
	j++;
     }

     return (i-att_rev);
}*/


void RevocaClassicLkh(nodosd *utenti, long nutenti, long da_rev)
{
	utenti[da_rev-nutenti+1].revocato=TRUE;
}

void AbilitaUtenteClassicLkh(nodosd *utenticlkh, long nutenti, long da_abilitare)
{
	//printf("\tLKH abilita %ld che diventa %ld\n",da_abilitare,da_abilitare+nutenti-1);
	utenticlkh[da_abilitare].revocato=FALSE;
	utenticlkh[da_abilitare].shortkeys=NULL;
}

int VerificaClassicLkh(chiave sess,nodosd *utenti,long nutenti)
{
	int i;
	listachiavi *l;

	for(i=0;i<nutenti;i++)
	{
		l=(utenti[i]).shortkeys;
		while(l!=NULL && memcmp(l->valore,sess,SHA_DIGEST_LENGTH)!=0)
		{
			l=l->next;
		}
		// se l'utente i-esimo ha la chiave ed e' revocato
		if( (l!=NULL && utenti[i].revocato) )
		{
			printf("\nL'utente %d ha la chiave ma e' revocato\n",i);
			//...c'e' un errore!
			return FALSE;

		}
		if( (l==NULL && !(utenti[i].revocato)))
		{
			printf("\nL'utente %d non ha la chiave ma e' legittimo\n",i);
			//...c'e' un errore!
			return FALSE;
		}
	}
	return TRUE;
}
