#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifndef TIPI
	#include "tipi.h"
#endif
#include "steiner.h"


typedef long* stree;

/*funzione che effettua una ricerca lineare dell'elemento su un array
di interi positivi, restituisce -1 se l'indice non  presente, altrimenti l'indice
della pasizione in cui si trova l'elemento*/
long search(stree st,long primo,long ultimo,long elemento){
	
	long i;
	
	i=primo;
	for (i=primo;i<ultimo+1;i++){
		if(st[i]==elemento){
			return i;
		}
	}
	return -1;
}

int NumeroSubset(long vecchionum, long nuovonum, long utenti_da_rev)
{
	if(nuovonum>(vecchionum+2*utenti_da_rev))
	{
		return FALSE;
	}
	return TRUE;
}

//funzione che preso un nodo restituisce 1 se il nodo  foglia, 0 altrimenti
int isfoglia(long nodo,long *arr,long lunghezza){
	
	if ((search(arr,0,lunghezza-1,nodo*2+1)==-1) && (search(arr,0,lunghezza-1,nodo*2+2)==-1)){ //cerco se nell'array c' uno dei due figli
	return 1;
	}
	else return 0;
}

//funzione che verifica se un dato figlio (lato=1 per destro, lato =0 per sinistro) di un nodo v esiste, intal caso viene restituito 1, altimenti 0
int esistefiglio(long v, stree arr,long lunghezza, int lato){
	
	if (lato==0){
		if (search(arr,0,lunghezza-1,v*2+1)==-1){
			return 0;
		}
	}
	else if (lato==1){
		if (search(arr,0,lunghezza-1,v*2+2)==-1){
			return 0;
		}
	}
	return 1;
}


//funzione che trova un sottoalbero con due foglie tale che la radice  il pi basso nodo possibile(spiegare meglio).
//NB: nel caso in cui l'albero degenerasse in pila, allora la funzione torna come risultato 1 e v=0
long trova2albero(long nodo,stree arr,long lunghezza,long *v){
	
	long fogliedx,fogliesx; //numero di foglie del sottoalbero destro e del sottoalbero sinistro
	
	fogliedx=0;fogliesx=0;
	if (isfoglia(nodo,arr,lunghezza)){
		return 1;
	}
	else{
		if (esistefiglio(nodo,arr,lunghezza,0)){ //se esiste il figlio sinistro
			fogliesx = trova2albero(nodo*2+1,arr,lunghezza,v); //calcola le foglie del sottoalbero sinistro
			if (fogliesx==1){
				if (esistefiglio(nodo,arr,lunghezza,1)){ //se esiste il figlio destro
					fogliedx = trova2albero(nodo*2+2,arr,lunghezza,v); //calcola le foglie del sottoalbero destro
				if ((fogliedx==1) && (*v==0)){
					*v=nodo;
					}
				}
			}
			return fogliedx+fogliesx;
		}
		else if (esistefiglio(nodo,arr,lunghezza,1)){ //se esiste il figlio destro
			fogliesx = trova2albero(nodo*2+2,arr,lunghezza,v);
			return fogliesx;
		}
		else{
			printf("Un caso di trova2albero non  stato considerato\n");
			exit(0);
		}
	}
}


//funzione che prende un albero binario radicato in nodo con due foglie, e restituisce in i e j (inizialmente posti a 0)i valori del nodo.
void fogliediv(long nodo,stree arr,long lunghezza,long *i,long *j){
	
	if (isfoglia(nodo,arr,lunghezza)){
		if (*i==0){
			*i=nodo;
		}
		else{
			*j=nodo;
		}
	}
	if(esistefiglio(nodo,arr,lunghezza,0)){ //se esiste il figlio sinistro
		fogliediv(nodo*2+1,arr,lunghezza,i,j); //lo cerco nel sottoalbero sx
	}
	if (esistefiglio(nodo,arr,lunghezza,1)){ //se esiste il figlio destro
		fogliediv(nodo*2+2,arr,lunghezza,i,j); //lo cerco nel sottoalbero dx
	}
}


//funzione che prende uno steiner tree e restituisce le foglie i,j (i  la foglia sinistra e j la sinistra) e il loro ultimo antenato comune v
//(secondo la notazione usata da Naor).
long trovaIJV(stree arr,long lunghezza,long *i,long *j){
	
	long v=0;
	
	*i=0;*j=0; //inizializzo le variabili, altrimenti la funzione fogliediv non funziona
	if (trova2albero(0,arr,lunghezza,&v)==1){ //se trova2albero torna 1 l'albero in verit era una lista
		*i=arr[lunghezza-1]; //essendo lo steiner tree ordinato, allora l'unica foglia  nell'ultimo elemento
		*j=*i;		//pongo Vi = Vj
	}
	else{
		fogliediv(v,arr,lunghezza,i,j); //fogliediv setta apportunamente Vi e Vj
	}
	return v;
}

//pone nell'array i figli di v pari a -1
void marcanodi(stree arr,long lunghezza,long nodo){
	
	long posizione;
	
	if (isfoglia(nodo,arr,lunghezza)){ //fino a che non trovo una foglia
		posizione= search(arr,0,lunghezza-1,nodo); //salvo l'indice della posizione nodo nel vettore dello steiner tree
		arr[posizione]=-1; //marco il nodo corrente
	}
	else{
		if(esistefiglio(nodo,arr,lunghezza,0)){ //se esiste il figlio sinistro
			marcanodi(arr,lunghezza,nodo*2+1); //marco il sottoalbero sinistro
		}
		if(esistefiglio(nodo,arr,lunghezza,1)){ //se esiste il figlio destro
			marcanodi(arr,lunghezza,nodo*2+2); //marco il sottoalbero destro
		}
		posizione= search(arr,0,lunghezza-1,nodo); //salvo l'indice della posizione nodo nel vettore dello steiner tree
		arr[posizione]=-1; //marco il nodo corrente
	}
}

//elimina il sottoalbero radicato in nodo
long eliminanodi(stree *arr,long lunghezza, long nodo){
	
	long i,j,contatore;
	
	i=0;j=0;contatore=0;
	marcanodi(*arr,lunghezza,nodo*2+1); //marco il sottoalbero sinistro
	marcanodi(*arr,lunghezza,nodo*2+2); //marco il sottoalbero destro
	 
	//compatto l'array;
	
	while((i<lunghezza) && (j<lunghezza)){
		while((i<lunghezza) && ((*arr)[i]!=-1)) i++;  //si posiziona sul primo elemento = -1
		j=i+1; //j  l'elemento subito dopo i
		while((j<lunghezza) && ((*arr)[j]==-1)) j++; ////si posiziona sul primo elemento != -1
		if ((i<lunghezza) && (j<lunghezza)){
			(*arr)[i]=(*arr)[j]; //scambio gli elementi
			(*arr)[j]=-1;
			j++;
			i++;
		}
	}
	while (((*arr)[contatore]!=-1)&&(contatore<lunghezza)) contatore++; //conta il numero di elementi diversi da -1
	if (!(*arr=realloc(*arr,contatore*sizeof(long)))){ //restrigo il vettore
		printf("\terrore nella realloc, funzione eliminanodi, ci sono %ld elementi nel vecchio vettore,si richiedono %ld elementi\n",lunghezza,contatore);
		exit(0);
	}
	return contatore; //torno la nuova dimensione dell'array
}

//funzione che prende uno steiner tree e restituisce un insieme di Sij nell'array copertura, restituisce il numero di elementi di copertura
long creasij(stree *arr, long lunghezza,sij **cover, long dimensionecopertura){
	
	sij elemento;
	long i,j,v,l,k;
	
	while (lunghezza>1){
		v=trovaIJV(*arr,lunghezza,&i,&j); //trovo la tripla v,i,j (v padre, i foglia sinistra , j foglia destra)
		if (i==j){ //lo steiner tree era una pila
			l=v;
			k=v; //pongo v=l=k
			elemento.i=l; 
			elemento.j=i;
			if(!(*cover=realloc(*cover,(dimensionecopertura+1)*sizeof(sij)))){
					printf("errore nella realloc, funzione creasij\n");
					exit(0);
				}
			dimensionecopertura++; //aumento il numero di elementi
			(*cover)[dimensionecopertura-1]=elemento; //assegno l'elemento all'array
			return dimensionecopertura;
		}
		else{
			l=v*2+1; //l  il figlio destro di v
			k=v*2+2; //k  il figlio sinistro di v
			if(l!=i){ //secondo le direttive di Naor
				elemento.i=l; 
				elemento.j=i;
				if(!(*cover=realloc(*cover,(dimensionecopertura+1)*sizeof(sij)))){
					printf("errore nella realloc, funzione creasij\n");
					exit(0);
				}
				dimensionecopertura++; //aumento il numero di elementi
				(*cover)[dimensionecopertura-1]=elemento; //assegno l'elemento all'array
			}
			if(k!=j){ //secondo le direttive di Naor
				elemento.i=k; 
				elemento.j=j;
				if(!(*cover=realloc(*cover,((dimensionecopertura+1)*sizeof(sij))))){
					printf("errore nella realloc, funzione creasij\n");
					exit(0);
				}
				dimensionecopertura++; //aumento il numero di elementi
				(*cover)[dimensionecopertura-1]=elemento; //assegno l'elemento all'array
			}
		}
		lunghezza=eliminanodi(arr,lunghezza,v); //elimino i discendenti di v
	}
	return dimensionecopertura;
}

//funzione che calcola la copertura Sij nell'array copertura e restituisce il numero si elementi della copertura
//copertura si assume pari a null
long calcolasij(sdtree alberosd, copertura *cover,l_revocati lista_revocati){
	
	long dimsteiner;
	long *stainertree;
	
	stainertree=NULL;
// 	printf("lista_revocati.tot_revocati vale: %ld\n",lista_revocati.tot_revocati);
// 	printf("(lista_revocati.revocati)[lista_revocati->tot_revocati] vale:%ld\n",(lista_revocati.revocati)[lista_revocati.tot_revocati-1]);
	//METTERE MEGLIO QUESTA RIGA
	dimsteiner=stainer_tree(&stainertree,alberosd.numerofoglie,alberosd.albero,lista_revocati.revocati,lista_revocati.tot_revocati); //calcolo lo steiner tree	
	if ((dimsteiner==0) || (stainertree==NULL)){
		printf("STEINER E' VUOTO\n");
		exit(0);
	}
// 	printf("Dimsteiner vale:%ld\n",dimsteiner);
	if (dimsteiner==1){ //se esiste 1 solo elemento allora i=0 j=unicorevocato
		if (!(cover->elementi=(realloc(cover->elementi,sizeof(sij))))){
			printf("Errore nell'allocazione della capoerura\n");
			exit(0);
		}
		(cover->elementi)->i=0;
		(cover->elementi)->j=stainertree[0];
		cover->dimensione++;
		return 1;
	}
	cover->dimensione=creasij(&stainertree,dimsteiner,&(cover->elementi),0); //altrimenti calcolo la copertura
//     printf("Dimensione della copertura: %ld\n",cover->dimensione);
	free(stainertree); //libero la memoria dello steiner tree
	return cover->dimensione;
	
	}
	
int comparatore(void *x,void *y){
	
	long *a,*b;
	
	a=(long *) x; b=(long *)y;
	if (*a<*b){
		return -1;
	}
	else if(*a>*b){
		return 1;
	}
	else return 0;
}


//funzione che controlla se la copertura trovata copre effettivamente tutti e soli i revocati, prende l'array di insiemi sij e la lista(ordinata) di nodi
//revocati, restituisce 0 se sij ricopre revocati, =! 0 altrimenti
int controlla_copertura(copertura cop,l_revocati list_revocati,sdtree *alberosd){
	
	long *appoggio; //array di appoggio in cui costruisco la copertura di Sij
	long abmin,abmax,rvmin,rvmax; //variabili che servono per mettere prima e ultima foglia dei revocati e abilitati Sij
	float altezzaj,altezzah;	//altezza nodo j e totale dell'albero
	char risultato=0;		//qui costruisco il risultato
	long i,j,h,k;	
	long ultimo=0;			//ultimo indice dell'array di appoggio
	long *l_abilitati;		//lista di utenti abilitati
	long numeroabilitati;			//numero di abilitati
	long dim_copertura;
	sij *cover;
	long *lista_revocati;
	long num_rev;
	long numero_utenti;
	
	numero_utenti=alberosd->numerofoglie;
	dim_copertura=cop.dimensione;
	cover=cop.elementi;
	lista_revocati=list_revocati.revocati;
	num_rev=list_revocati.tot_revocati;
	numeroabilitati=numero_utenti-num_rev;  //il numerodegliabilitati  il totale meno i revocati
	
	if (!(appoggio = calloc(numeroabilitati,sizeof(long)))){; //alloco lo spazio per gli abilitati
		printf("\tErrore nell'allocazione in controlla copertura di %ld elementi\n",numeroabilitati);
		exit(0);
	}
	modff((float)(log10(numero_utenti)/log10(2)),(float *)&altezzah); //calcolo l'altezzadell'albero completo
	for (k=0;k<dim_copertura;k++){
		j=cover[k].j;	//in j metto il nodo che sottende i revocati
		modff(log10((float)(j+1))/log10(2),(float *)&altezzaj);	//calcolo l'altezza del sottoalbero radicati in j
		rvmin=pow(2,altezzah-altezzaj)*j+pow(2,altezzah-altezzaj)-1; //calcolo la foglia pi a sinistra
		rvmax=pow(2,altezzah-altezzaj)*j+pow(2,(altezzah-altezzaj)+1)-2; //calcolo la foglia pi a sinistra
		i=cover[k].i;	//in i metto il nodo che sottende gli abilitati
		modff(log10((float)(i+1))/log10(2),(float *)&altezzaj);	//calcolo l'altezza del sottoalbero radicati in i
		abmin=pow(2,altezzah-altezzaj)*i+pow(2,altezzah-altezzaj)-1; //calcolo la foglia pi a sinistra
		abmax=pow(2,altezzah-altezzaj)*i+pow(2,(altezzah-altezzaj)+1)-2; //calcolo la foglia pi a sinistra
		h=abmin;
		while (h<=abmax){
			if (ultimo>numeroabilitati){ //la copertura ha superrato gli abilitati reali
				printf("\tLa copertura trovata  troppo grande\n");
				printf("\tnumeroabilitati vale: %ld, ultimo: %ld\n",numeroabilitati,ultimo);
				printf("\tGli abilitati sono:\n\t");
				for(h=0;h<numeroabilitati;h++){
					printf("%ld\t",appoggio[h]);
				}
				printf("\n");
				return 1;
			}
			if((h<rvmin)||(h>rvmax)){
				appoggio[ultimo]=h;
				ultimo++;
			}
			h++;
		}
	}
	if (ultimo!=numeroabilitati){ //se i due array non hanno gli stessi elementi videntemente la copertura  errata  
		printf("\tLe coperture non coincidono: ultimo=%ld\t,n abilitati=%ld\n",ultimo,numeroabilitati);
		printf("\tGli abilitati sono:\n\t");
				for(h=0;h<numeroabilitati;h++){
				printf("%ld\t",appoggio[h]);
			}
		printf("\n");
		return 1;
	}	
	qsort(appoggio,numeroabilitati,sizeof(long),comparatore); //ordino l'array
	
	//alloco e riempio la lista degli abilitati
	if (!(l_abilitati = calloc(numeroabilitati,sizeof(long)))){; //alloco lo spazio per gli abilitati
		printf("\tErrore nell'allocazione in controlla copertura di %ld elementi per l_abilitati\n",numeroabilitati);
		exit(0);
	}
// 	printf("In revoca:\n");
//  	printf("num_rev: %ld\n",num_rev);
// 	printf("num abilitati: %ld\n",numeroabilitati);
 	abilitati(l_abilitati,alberosd->albero,alberosd->numerofoglie);
	for (i=0;i<numeroabilitati;i++){
		l_abilitati[i]+=alberosd->numerofoglie-1;
	}
// 	printf("l_abilitati:\t");
//  	for(h=0;h<numeroabilitati;h++)
//  		{
//  				printf("%ld ",l_abilitati[h]);
//  		}
//  		printf("\n");
// 	printf("appoggio:\t");
//  	for(h=0;h<numeroabilitati;h++)
//  		{
//  				printf("%ld ",appoggio[h]);
//  		}
//  		printf("\n");
	for (h=0;h<numeroabilitati;h++){
		risultato= (l_abilitati[h]^appoggio[h]) | risultato;
	}
	free(appoggio);
	free(l_abilitati);
	return risultato;
	}
