#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include "lkh.h"
#include "tree.h"
#include "key.h"
#include "lkhutil.h"


/* Init(): inizializza la struttura puntata da <centralelkh> con i dati presi come argomenti, e crea la radice dell'albero lkh la cui testa viene posta nel campo root di <centralelkh>*/
int initlkh(lkh *centralelkh, int maxfigli, int sbilanciamento)
{
	alberolkh *radice;

	/* i campi d e c di <centralelkh> sono inizializzati con i dati specificati dal chiamante, il campo subsets viene posto a NULL, poiche' non e' stata eseguita alcuna revoca in SD per ora */
	centralelkh->d=maxfigli;
	centralelkh->c=sbilanciamento;
	centralelkh->subsets=NULL;
	centralelkh->nsubsets=0;

	/* alloca la radice dell'albero dell'lkh controlla che l'allocazione vada a buon fine*/
	if((radice=malloc(sizeof(alberolkh)))==NULL)
	{
		free(centralelkh);
		return FALSE;
	}

	/* inizializza opportunamente i campi della radice, se ci sono errori di allocazione dealloca tutto cio' che e' stato precedentemente allocato ed esce restituendo al chiamante FALSE */
	if((radice->sons=calloc(maxfigli,sizeof(alberolkh *)))==NULL)
	{
		free(radice);
		return FALSE;
	}
	radice->nfigli=0; /* la radice per ora non ha figli*/
	radice->mysubset=NULL; /* la radice e' un nodo interno, dunque non ha subset */
	/* la chiave non viene inizializzata poiche' al primo inserimento nell'albero la chiave della radice dovra' essere riassegnata*/
	radice->iscorrupted=FALSE; /* il path cui appartiene la radice inizialmente non e' corrotto */
	radice->parent=NULL; /* la radice in quanto tale non ha alcun genitore */

	centralelkh->root=radice;




	return TRUE;
}


/* Lkh(): implementa la variante dell'LKH applicata nell'algoritmo SD+LKH. Prende in input una struttura lkh, la lista dei nuovi subsets e il loro numero e si preoccupa di aggiornare opportunamente l'albero lkh (inserendo i nuovi subsets, cancellando quelli non piu' presenti ed eventualmente operandone il bilanciamento) e di operare il rekeying dell'albero. Restituisce il numero di messaggi impiegati per effettuare l'operazione di rekeying dell'albero. */
long Lkh(lkh *centrale, Sij *attuali, long nsubsets, nodosd *utentisd, long nutentisd)
{
	Sij *dainserire;
	long nins,ncan;
	long i;
	Sij *dacancellare;
	long x;
	FILE *f;

	f=fopen("nsubset.dat","a");
	fprintf(f,"%ld\n",nsubsets);
	fclose(f);

	
	//printf("Subset da processare: ");
	f=fopen("subset.dat","a");
 	for(i=0;i<nsubsets;i++)
 	{
 		//printf("(%ld,%ld) ",attuali[i].i,attuali[i].j);
 		fprintf(f,"(%ld,%ld) ",attuali[i].i,attuali[i].j);
 	}
	fprintf(f,"\n");
// 	printf("\n");
	fclose(f);






	ncan=0;
	
// 	printf("\n0Subset da processare: ");
// 	for(i=0;i<nsubsets;i++)
// 	{
// 		printf("(%ld,%ld) ",attuali[i].i,attuali[i].j);
// 	}
// 	printf("\n");
	
	/*printf("\nSubset vecchi: ");
	for(i=0;i<centrale->nsubsets;i++)
	{
		printf("(%ld,%ld) ",(centrale->subsets)[i].i,(centrale->subsets)[i].j);
	}
	printf("\n");*/
	


	/* pone in <dainserire> tutti gli Sij che appartengono ad <attuali> ma non a <centrale->subsets> (ovvero tutti gli Sij che vanno inseriti nell'albero LKH) e in <nins> il loro numero. In caso di errore esce restituisce al chiamante ERRORE */

	nins=Differenza(attuali,centrale->subsets,nsubsets,centrale->nsubsets,&dainserire);
	if(nins==ERRORE)
	{
		return ERRORE;
	}


	/* analogamente, pone in <dacancellare> tutti i nodi che appartengono a <centrale->subsets> ma non a <attuali> (ovvero i nodi che vanno eliminati dall'albero LKH perche' corrispondenti a subset revocati) e in <ncan> il loro numero */

// 	printf("\n1Subset da processare: ");
// 	for(i=0;i<nsubsets;i++)
// 	{
// 		printf("(%ld,%ld) ",attuali[i].i,attuali[i].j);
// 	}
// 	printf("\n");
	
	ncan=Differenza(centrale->subsets,attuali,centrale->nsubsets,nsubsets,&dacancellare);
	if(ncan==ERRORE)
	{
		if(nins)
		{
			free(dainserire);
		}
		return ERRORE;
	}

	
	f=fopen("ins.dat","a");
// 	printf("Da inserire: ");
// 	for(i=0;i<nins;i++)
// 	{
// 		printf("(%ld,%ld) ",dainserire[i].i,dainserire[i].j);
// 	}
// 	printf("\n");
	fprintf(f,"%ld\n",nins);
	fclose(f);

	f=fopen("canc.dat","a");
// 	printf("Da cancellare: ");
// 	for(i=0;i<ncan;i++)
// 	{
// 		printf("(%ld,%ld) ",dacancellare[i].i,dacancellare[i].j);
// 	}
// 	printf("\n");
	fprintf(f,"%ld\n",ncan);
	fclose(f);
	
// 	printf("\n2Subset da processare: ");
// 	for(i=0;i<nsubsets;i++)
// 	{
// 		printf("(%ld,%ld) ",attuali[i].i,attuali[i].j);
// 	}
// 	printf("\n");
	



	/* privilegia la sostituzione rispetto all'inserimento e alla cancellazione dei subsets: sostituisce il numero maggiore di subsets possibili, ovvero il minimo tra <ninserimenti> ed <ncancellazioni>. Alla fine del ciclo, ovviamente, saranno rimasti o soltanto nodi da inserire o soltanto nodi da cancellare, rispettivamente se (ninserimenti>ncancellazioni) oppure se (ncancellazioni>ninserimenti) */
	for(i=0;i<Min(nins,ncan);i++)
	{
// 		printf("Sostituisco a (%ld,%ld) il subset (%ld,%ld)\n",dacancellare[i].i,dacancellare[i].i,dainserire[i].i,dainserire[i].j);
		if(Substitute(centrale->root,dacancellare[i],dainserire[i])==FALSE)
		{
			return ERRORE;
		}
	}


	/* poiche' come detto al termine dei cicli precendenti sono rimaste solo cancellazioni da effettuare oppure soltanto inserimenti, ovvero l'indice i e' pari a <ncancellazioni> oppure a <ninserimenti>, segue che soltanto uno dei due cicli sottostanti verra' eseguito realmente*/

	/* finche' le cancellazioni da effettuare non sono terminate...*/
	while(i<ncan)
	{
		/* cancella il subset i-esimo della lista dei subset da eliminare*/
		if(Delete(centrale->root,dacancellare[i])==FALSE)
		{
			return ERRORE;
		}
		i++;
	}

	/* finche' gli inserimenti da effettuare non sono terminati...*/
	while(i<nins)
	{
		/*...inserisci il subset i-esimo della lista dei subset da inserire*/
		if(Inserisci(centrale->root,dainserire[i],centrale->d)==FALSE)
		{
			return ERRORE;
		}
		i++;
	}
	
	/* prima di bilanciare simula l'esecuzione del Rekeying, in modo da separare i messaggi che dipendono dal bilanciamento da quelli che dipendono dall'algoritmo vero e proprio */
	 x=SimulaRekeying(centrale->root, utentisd, nutentisd,0);
	 
	 	
	/* a seguito degli inserimenti o delle cancellazioni effettuate, l'albero potrebbe essere sbilanciato, chiama dunque Balance() che provvedera' a controllare il bilanciamento, e nel caso a bilanciare l'albero */

	if((i=Balance(centrale->root,centrale->c,centrale->d))==FALSE)
	{
		return ERRORE;
	}
	/*else if(i==2)
	{
		printf("B");
	}*/

	
	/* infine, dopo aver ottenuto un albero aggiornato bilanciato, chiama la funzione Rekeying, che si preoccupa di riassegnare le chiavi dei path corrotti durante le operazioni di aggiornamento. Rekeying restituisce il numero di messaggi spediti nel corso di tale operazione, e questo e' il dato che la funzione Lkh restituisce al chiamante. */

	 if((i=Rekeying(centrale->root, utentisd, nutentisd,0))==ERRORE)
	 {
	 	printf("Errore su Rekeying!\n");
	 	return ERRORE;
	 }
	 
	 /* salva su file i messaggi che dipendono dal bilanciamento */
	if((f=fopen("bil_hsd.dat","a"))==NULL)
	{
		printf("ERRORE impossibile salvare i dati relativi al bilanciamento\n");
		return ERRORE;
	}
	
	fprintf(f,"%ld\n",i); /* il numero di mex dovuti al bilanciamento e' pari al numero di mex totali meno quelli dovuti esclusivamente all'algoritmo senza bilanciamento.*/
	fclose(f);

	 
	  /* Poiche' <centrale> non contiene una lista di subsets aggiornati, elimina la lista corrente e la sostituisce con la lista attuale */
	if(centrale->subsets)
	{
		free(centrale->subsets);
	}


	/*centrale->subsets=attuali;
	centrale->nsubsets=nsubsets;*/

	/* libera la memoria occupata dalla lista dei subsets da inserire e da cancellare */
	free(dainserire);
	free(dacancellare);
	
	centrale->subsets=calloc(nsubsets,(sizeof(sij)));
	for(i=0;i<nsubsets;i++) (centrale->subsets)[i]=attuali[i];
	centrale->nsubsets=nsubsets;

	return x;
}

/* Sessionkey(): restituisce la chiave della sessione attuale */
void Sessionkey(lkh centrale, chiave disessione)
{
	memcpy(disessione,(centrale.root)->key,SHA_DIGEST_LENGTH);
}
