// #include <openssl/sha.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include "tree.h"
#include "key.h"
#include "communication.h"
#include "lkhutil.h"

long PrimoAntenatoCorrotto(alberolkh *nodo);


/*Corrompi(): corrompe il path da <nodo> fino alla radice dell'albero lkh*/
void Corrompi(alberolkh *nodo)
{
	alberolkh *dacorrompere;

	dacorrompere=nodo;
	/* finche' non e' giunto alla radice... */
	while(dacorrompere!=NULL)
	{
		/* corrompe il nodo corrente */
		dacorrompere->iscorrupted=TRUE;
		/* passa al padre del nodo corrente */
		dacorrompere=dacorrompere->parent;
	}
}

/*IsLeaf(): torna TRUE se <nodo> e' una foglia, FALSE altrimenti */
int IsLeaf(alberolkh *nodo)
{
	if(!nodo) return FALSE;
	return (nodo->nfigli==0) ? TRUE : FALSE; /* torna ok se il nodo ha zero figli (ovvero e'una foglia*/
}

/* MinDeep(): ritorna l'altezza del sottoalbero di <radice> avente altezza minore */
int MinDeep(alberolkh *radice)
{
	int minaltezza;
	int attuale;
	int i;

	if(radice==NULL)
	{
		return 0;
	}
	else if(IsLeaf(radice))
	{
		return 1;
	}
	else
	{
		minaltezza=Deep(radice); /*inizializza la minima altezza come altezza della radice stessa (in tal modo ogni figlio avra' altezza minore di minaltezza cosi' inizializzata)*/
		for(i=0;i<radice->nfigli;i++)
		{
			attuale=MinDeep((radice->sons)[i]);
			if(attuale<minaltezza) minaltezza=attuale;
		}
	}
	return (minaltezza+1);
}

/* Deep(): ritorna l'altezza dell'albero radicato in <radice> */
int Deep(alberolkh *radice)
{
	int maxaltezza;
	int attuale;
	int i;

	if(radice==NULL)
	{
		return 0;
	}
	else if(IsLeaf(radice))
	{
		return 1;
	}
	else
	{
		maxaltezza=0; /*inizializza la massima altezza a zero (esistera' almeno un figlio con altezza maggiore, essendo <radice> non foglia)*/
		for(i=0;i<radice->nfigli;i++)
		{
			attuale=Deep((radice->sons)[i]);
			if(attuale>maxaltezza) maxaltezza=attuale;
		}
	}
	return (maxaltezza+1);
}

/* IsBalanced(): restituisce TRUE se l'albero radicato in <radice> e' bilanciato, FALSE altrimenti */
int IsBalanced(alberolkh *radice, int sbilanciamento, int maxfigli)
{
	int livmin,livmax,nfratelli;

	LastLeaf(radice,&livmin,maxfigli);
	MaxLeaf(radice,&livmax,&nfratelli);
	/* Se la differenza tra l'altezza dell'albero (ovvero il livello della foglia piu' "bassa" dell'albero) e l'altezza della foglia che ha livello minore tra tutte quelle dell'albero (ovvero la piu' "alta" dell'albero) e' maggiore del fattore di sbilanciamento, torna FALSE poiche' l'albero non e' bilanciato. In caso contrario, poiche' siamo di fronte ad un albero bilanciato, torna TRUE.*/
	
	//if ((livmax-livmin)>sbilanciamento) printf("\n Foglia MAX %d Foglia MIN %d Sbilanciato %d\n", livmax, livmin, livmax-livmin);
	
	return ((livmax-livmin)>sbilanciamento) ? FALSE : TRUE;
}

/*LastLeaf(): ritorna il puntatore alla foglia di altezza minore dell'albero radicato in <radice>*/
alberolkh *LastLeaf(alberolkh *radice, int *altezzafoglia,int maxfigli)
{
	int minaltezza;
	alberolkh *fogliamin=NULL;
	alberolkh *fogliaatt;
	int attuale;
	int i;


	if(radice==NULL)
	{
		*altezzafoglia=0;
		return NULL;
	}
	else if(IsLeaf(radice))
	{
		*altezzafoglia=1;
		return radice;
	}
	else
	{
		/* se il nodo non e' al completo, e' possibile inserire al suo interno un figlio, dunque il sottoalbero minore radicato in <radice> ha altezza 1 (e coincide con la radice stessa) */
		if(radice->nfigli<maxfigli)
		{
			*altezzafoglia=1;
			return radice;
		}
		minaltezza=Deep(radice); /*inizializza la minima altezza come altezza della radice stessa (in tal modo ogni figlio avra' altezza minore di <minaltezza> cosi' inizializzata)*/
		for(i=0;i<radice->nfigli;i++)
		{
			fogliaatt=LastLeaf((radice->sons)[i],&attuale,maxfigli);
   			if(attuale<minaltezza) /* se l'altezza di <fogliaatt> e' minore dell'altezza della foglia piu' bassa finora trovata...*/
			{
				/*... <fogliatt> diventa la foglia piu' bassa */
				 minaltezza=attuale;
				 fogliamin=fogliaatt;
			}
			/* altrimenti, se le altezze sono uguali, ma (radice->sons)[i] e' una foglia, prediligi l'inserimento in quest'ultima anziche' in un nodo con gia' almeno due figli*/
			else if(attuale==minaltezza && IsLeaf((radice->sons)[i]))
			{
				minaltezza=attuale;
				fogliamin=fogliaatt;
			}
		}
	}
	*altezzafoglia=minaltezza+1;

	return fogliamin;
}

/*MaxLeaf(): ritorna il puntatore alla foglia di altezza massima (a parita' di altezza si considera prima la prossimita' ad un antenato corrotto e poi il numero di fratelli)  dell'albero radicato in <radice>*/
alberolkh *MaxLeaf(alberolkh *radice, int *altezzafoglia, int *nfratelli)
{
	int maxaltezza;
	alberolkh *fogliamax;
	alberolkh *fogliaatt;
	int attuale,maxfratelli,fratelli;
	int i;
	long dist1,dist2;

	fogliamax=NULL;

	if(radice==NULL)
	{
		*altezzafoglia=0;
		*nfratelli=0;
		return NULL;
	}
	else if(IsLeaf(radice))
	{
		*altezzafoglia=1;
		*nfratelli=0;
		return radice;
	}
	else
	{
		maxaltezza=MinDeep(radice)-1; /*inizializza la massima altezza come altezza del sottoalbero di <radice> di altezza minore (in tal modo esistera' almeno un figlio di altezza minore )*/
		maxfratelli=-1;
		for(i=0;i<radice->nfigli;i++)
		{
			fogliaatt=MaxLeaf((radice->sons)[i],&attuale,&fratelli);
			/* se l'altezza di <fogliaatt> e' minore dell'altezza della foglia piu' bassa finora trovata, oppure le altezze sono uguali ma <fogliaatt> ha piu' fratelli...*/
			/* se la foglia restituita da MaxLeaf e' una figlia del nodo corrente...*/
			if(attuale==1)
			{
				/*...allora i fratelli sono quanti i figli del nodo corrente meno uno*/
				fratelli=radice->nfigli-1;
			}
		
			if(fogliamax==NULL)
			{
				 maxaltezza=attuale+1;
				 fogliamax=fogliaatt;
				 maxfratelli=fratelli;
			}
			else
			{
				/* calcola le distanze di fogliamax e fogliaatt dal loro antenato corrotto piu' prossimo*/			
			
				dist1=PrimoAntenatoCorrotto(fogliaatt);
				dist2=PrimoAntenatoCorrotto(fogliamax);
			
				/* se la foglia trovata e' piu' in profondita' di quella trovata in precedenza, oppure e' alla stessa profondita' ma e' piu' vicina ad un nodo corrotto della precedente (e dunque costerebbe meno messaggi spostarla rispetto alla precedente), oppure e' alla stessa profondita' e ha la stessa distanza da un antenato revocato della precedente ma ha piu' fratelli...*/
   				if((attuale+1)>maxaltezza || ((attuale+1)==maxaltezza && dist1<dist2) || ((attuale+1)==maxaltezza && dist1==dist2 && fratelli>maxfratelli))
				{
					/*... <fogliatt> diventa la foglia piu' bassa */
				 	maxaltezza=attuale+1;
				 	fogliamax=fogliaatt;
				 	maxfratelli=fratelli;
				}
			}
		}
	}
	*altezzafoglia=maxaltezza;
	*nfratelli=maxfratelli;

	return fogliamax;
}

/* Substitute(): sostituisce, all'interno dell'albero radicato in <radice>, la foglia con subset pari a <vecchio> con una foglia con subset pari a <nuovo> */
int Substitute(alberolkh *radice, Sij vecchio, Sij nuovo)
{
	int i;

	/* se la radice e' NULL...*/
	if(radice==NULL)
	{
		/*..ritorna FALSE, poiche' non c'e' l'Sij da sostituire*/
		return FALSE;
	}
	/*altrimenti, se <radice> e' una foglia...*/
	else if(IsLeaf(radice))
	{
		/*... se il subset di radice e' proprio quello che va sostituito...*/
		if((radice->mysubset)->i==vecchio.i && (radice->mysubset)->j==vecchio.j)
		{
			/*...sostituisce il vecchio subset con il nuovo */
			*(radice->mysubset)=nuovo;
			/* segna il nodo come corrotto */
			radice->iscorrupted=TRUE;
			/* corrompe il path dal nodo alla radice dell'albero lkh */
			Corrompi(radice);
			/* restituisce TRUE per segnalare l'avvenuta sostituzione */
			return TRUE;
		}
		/*...se il subset NON e' quello da sostituire...*/
		else
		{
			/* restituisce FALSE, per segnalare che l'Sij non e' stato trovato nel sottoalbero corrente*/

			return FALSE;
		}
	}
	/* se invece <radice> e' un nodo interno...*/
	else
	{
		/*...per ogni figlio di <radice>...*/
		for(i=0;i<radice->nfigli;i++)
		{
			/* se nel sottoalbero radicato nell'i-esimo figlio di <radice> la sostituzione ha esito positivo, allora restituisce TRUE per segnalare l'avvenuta sostituzione*/
			if(Substitute((radice->sons)[i],vecchio,nuovo)==TRUE)
			{
				return TRUE;
			}
		}
		/* se il ciclo e' terminato senza aver trovato l'Sij da sostituire, restituisce FALSE per segnalare che l'Sij in questione non e' presente nel sottoalbero corrente */
		return FALSE;
	}

}

/* Delete(): cancella la foglia con Sij pari a <dacancellare>, se questa e' presente all'interno dell'albero radicato in <radice>. Torna TRUE in caso di successo, FALSE altrimenti*/
int Delete(alberolkh *radice, Sij dacancellare)
{
	int i;
	alberolkh *padre,*figlio;

	/* se la radice e' NULL...*/
	if(radice==NULL)
	{
		/*..ritorna FALSE, poiche' non c'e' nodo da cancellare*/
		return FALSE;
	}
	/*altrimenti, se <radice> e' una foglia...*/
	else if(IsLeaf(radice))
	{
		/*... se il subset di radice e' proprio quello che va cancellato...*/
		if((radice->mysubset)->i==dacancellare.i && (radice->mysubset)->j==dacancellare.j)
		{
			padre=radice->parent;
			/*...cerca il puntatore che lega il padre alla foglia da cancellare */
			for(i=0;i<padre->nfigli && (padre->sons)[i]!=radice;i++);
			/* alla fine del ciclo, in (padre->sons)[i] c'e' il puntatore al nodo da cancellare*/
			(padre->sons)[i]=(padre->sons)[padre->nfigli-1]; /* sovrascrive il puntatore al nodo da cancellare con il puntatore all'ultimo dei figli di padre */

			/* dealloca il nodo radice */
			free(radice->sons);
			free(radice->mysubset);
			free(radice);
			
			padre->iscorrupted=TRUE;
			/* corrompe il path dal nodo <padre> alla radice dell'albero lkh */
			Corrompi(padre);

			/* decrementa di una unita' il numero di figli di <padre>*/
			(padre->nfigli)--;
			/* se al padre e' rimasto un solo figlio, e il padre non e' la radice dell'albero LKH (nel qual caso esso deve rimanere invariato)...*/
			if(padre->nfigli==1 && padre->parent!=NULL)
			{
				/*...il figlio viene sostituito al proprio genitore*/
				figlio=(padre->sons)[0];
				//memcpy(padre->key,figlio->key,SHA_DIGEST_LENGTH);
				padre->mysubset=figlio->mysubset;
				free(padre->sons);
				padre->sons=figlio->sons;
				padre->nfigli=figlio->nfigli;
				padre->iscorrupted=figlio->iscorrupted;
				/* e la foglia che rappresentava il figlio viene cancellata*/
				free(figlio);
				for(i=0;i<padre->nfigli;i++)
				{
					figlio=(padre->sons)[i];
					figlio->parent=padre;
				}

			}

			/* restituisce TRUE, poiche' la foglia cercata e' stata trovata e cancellata */
			return TRUE;
		}
		/* altrimenti, se la foglia in questione NON e' l'insieme Sij desiderato...*/
		else
		{
			/*...restituisce FALSE, poiche' in questo sottoalbero non e' presente tale insieme*/
			return FALSE;
		}
	}
	/* se invece <radice> e' un nodo interno, allora bisogna scandire tutti i sottoalberi radicati in ognuno dei suoi figli, per vedere se uno di essi contiene l'insieme Sij cercato*/
	else
	{
		/*per ogni figlio di <radice>...*/
		for(i=0;i<radice->nfigli;i++)
		{
			/*...se nel sottoalbero radicato nell'i-esimo figlio di <radice> la cancellazione ha esito positivo...*/

			if(Delete((radice->sons)[i],dacancellare)==TRUE)
			{
				/*...restituisce TRUE per segnalare l'avvenuta cancellazione*/
				return TRUE;
			}
		}
		/* se il ciclo e' terminato senza aver trovato l'Sij da cancellare, restituisce FALSE per segnalare che l'Sij in questione non e' presente nel sottoalbero corrente */
		return FALSE;
	}

}

/*PrimoAntenatoCorrotto(): ritorna il numero di nodi che separano <nodo> dall'antenato corrotto piu' vicino, oppure ERRORE se <nodo> non ha antenati corrotti */
long PrimoAntenatoCorrotto(alberolkh *nodo)
{
	long i;
	alberolkh *antenato;
	
	
	if(nodo==NULL)
	{
		return 0;
	}
	else if(nodo->parent==NULL)
	{
		if(nodo->iscorrupted) return 0;
		else return 1;
	}
	
	antenato=nodo;
	i=0;
	/* finche' non e' terminata la scansione all'indietro dell'albero (ovvero dopo aver scandito la radice), e finche' non si e' trovato un antenato corrotto... */
	while(antenato!=NULL && !antenato->iscorrupted)
	{
		i++;
		antenato=antenato->parent;
	}

	if(antenato==NULL)
	{
		return ERRORE;
	}
	else
	{
		return i;
	}
}

/*

se LastLEaf restituisce la radice, inserisci nel primo figlio libero della radice, altrimenti
se c'e' spazio inserisci nel padre della foglia restituita da LAstLeaf, se non c'e' spazio inserisci nella foglia di lastleaf
*/

alberolkh *BestJoin(alberolkh *radice, int livello, int maxfigli)
{
	alberolkh *nodoatt,*nodomigliore;
	long dist1,dist2;
	int i;


	/* se la radice non esiste, torna NULL poiche' non esiste alcun nodo di livello <livello> nel sottoalbero radicato in <radice>*/
	if(!radice)
	{
		return NULL;
	}
	/* se la radice e' una foglia, ovvero se il sottoalbero e' terminato...*/
	if (IsLeaf(radice))
	{
		/*...se il livello e' inferiore di quello richiesto...*/
		if(livello!=1)
		{
			/*...restituisce NULL per segnalare la situazione*/
			return NULL;
		}
		/* ...altrimenti restituisce l'unico nodo del sottoalbero radicato in <radice> di livello desiderato: ovvero proprio radice, poiche' questa e' una foglia*/
		else
		{
			return radice;
		}
	}
	/*...altrimenti, se e' possibile continuare a scorrere il sottoalbero...*/
	else
	{
		/* ...se <radice> e' al livello desiderato...*/
		if(livello==1)
		{
			/*...se e' possibile inserire all'interno di radice...*/
			if(radice->nfigli<maxfigli)
			{
				/*...restituisce al chiamante radice*/
				return radice;
			}
			/*...altrimenti, poiche' in questo sottoalbero non ci sono nodi di livello <livello> in cui e' possibile inserire...*/
			else
			{
				/*...restituisce NULL */
				return NULL;
			}
		}
		/*...se <radice> non e' al livello desiderato (ovvero se e' di altezza maggiore),  occorre scorrere tutti i suoi figli per scegliere il migliore dei nodi a livello <livello> in cui inserire (per migliore si intende un nodo corrotto, meglio se foglia)...*/
		else
		{

			/* inizializza nodomigliore a NULL, poiche' per ora non e' stato trovato alcun nodo di livello desiderato */
			nodomigliore=NULL;

			for(i=0;i<radice->nfigli;i++)
			{
				/*...cerca il miglior nodo al livello desiderato tra i nodi dell'albero radicato nell'i-esimo figlio di <radice> */
				nodoatt=BestJoin((radice->sons)[i],livello-1,maxfigli);
				/* se in tale sottoalbero esiste un nodo di livello desiderato in cui e' possibile inserire...*/
				if(nodoatt)
				{
					/*... e se non e' stato trovato precedentemente nessun nodo con le caratteristiche richieste...*/
					if(!nodomigliore)
					{
						/*...allora il nodo migliore sinora trovato e' proprio quello che BestJoin ha appena restituito, ovvero <nodoatt> */
						nodomigliore=nodoatt;
					}
					/*...altrimenti, se e' stato gia' selezionato un nodo ottimale per l'inserimento...*/
					else
					{
					/*...verifica se il nodo appena trovato e' preferibile rispetto al nodo che finora era considetato il migliore per il prossimo inserimento (perche' un nodo sia quello migliore, deve essere corrotto e deve avere il numero minimo di figli; a parita' di condizioni, si preferisce mantenere il nodo che risultava essere il migliore in precedenza) */
						dist1=PrimoAntenatoCorrotto(nodoatt);
						dist2=PrimoAntenatoCorrotto(nodomigliore);
						/* se il nodo appena trovato si trova pou' vicino a un path gia' corrotto rispetto al nodo che finora risulta migliore...*/
						if(dist1<dist2)
						{
							/*...allora il nodo attuale e' preferibile rispetto al migliore!*/
							nodomigliore=nodoatt;
						}
						/*altrimenti, se sono a pari distanza da un path corrotto, ma il nodo appena trovato ha meno figli del nodo ritenuto migliore...*/
							else
						if((dist1==dist2) && (nodoatt->nfigli)<(nodomigliore->nfigli))
							{
								/*...il nodo appena trovato e' da preferirsi al precedente */
								nodomigliore=nodoatt;
							}
					}

				}
			}
		/* al termine del for, dopo aver scandito tutti i sottoalberi radicati in ogni figlio di <radice>, restituisce al chiamante il nodo che e' risultato migliore */
		return nodomigliore;
		}
	}
}


/* Inserisci(): inserisce una foglia nell'albero radicato in <radice>, facendo in modo che il nodo inserito sia fratello della foglia di altezza minore oppure figlio di questa (in questo modo, se l'albero era bilanciato prima dell'inserimento, lo sara' anche dopo, poiche' lo sbilanciamento si sara' mantenuto inalterato se il nuovo nodo sara' stato inserito come fratello della foglia di altezza minore, o sara' diminuito di uno nell'altro caso) */
int Inserisci(alberolkh *radice, Sij insieme, int maxfigli)
{
	alberolkh *joinpoint;
	alberolkh *nuovo,*padre,*sostituto;
	int livellojp,i;


	/* trova il nodo di altezza minore in cui si puo' inserire */
	joinpoint=LastLeaf(radice,&livellojp,maxfigli);

	/* cerca, tra tutti i nodi di altezza pari a <livellojp> (ovvero l'altezza del nodo di altezza minore tra quelli in cui e' possibile inserire), il miglior candidato all'inserimento e lo pone in <joinpoint>*/
	joinpoint=BestJoin(radice, livellojp, maxfigli);

	if(livellojp==1) /* se l'albero e' costituito dalla sola radice...*/
	{
		/*...allora inserisce nel primo figlio libero della radice */
		padre=radice;
	}
	/* altrimenti, se joinpoint e' un nodo interno, ed e' possibile inserire dunque la foglia da creare come figlio di quest'ultimo...*/
	else if(!IsLeaf(joinpoint))
	{
		/* allora il padre della nuova foglia sara' proprio <joininpoint>*/
		padre=joinpoint;

	}
	/* altrimenti, se il nodo di altezza minore in cui inserire e' una foglia...*/
	else
	{
		/*...se c'e' la possibilita' di inserire il nuovo nodo come fratello della foglia piu' bassa, allora padre viene posto come il genitore della foglia restituita da LastLeaf */
		padre=joinpoint->parent;
		/*...altrimenti, se non ci sono posti disponibili, si inserisce il nuovo nodo come figlio della foglia <joinpoint> restituita da LastLeaf(): e poiche' non possono esserci nodi con un singolo figlio, un nuovo nodo,<sostituto> (senza subset, essendo un nodo interno) viene sostituito a <joinpoint>, che diventa uno dei due figli di tale nodo, insieme con il nodo (foglia) che il chiamante ha chiesto di inserire e di cui ha fornito i dati*/
		if(padre->nfigli>=maxfigli)
		{
			/* alloca il nodo che sostituira' <joinpoint> e controlla che l'allocazione vada a buon fine*/
			if((sostituto=malloc(sizeof(alberolkh)))==NULL)
			{
				return FALSE;
			}
			/* alloca i campi del nuovo nodo, se ci sono errori di allocazione 	dealloca tutto cio' che e' stato precedentemente allocato ed esce restituendo al chiamante FALSE */
			if((sostituto->sons=calloc(maxfigli,sizeof(alberolkh *)))==NULL)
			{
				free(sostituto);
				return FALSE;
			}
			/* la chiave di questo nodo non viene inizializzata, poiche' sara' assegnata all'atto del rekeying*/


			/* il subset di <sostituto> non esiste, essendo un nodo interno*/
			sostituto->mysubset=NULL;
			sostituto->iscorrupted=TRUE; /*ovviamente il path a cui appartiene il nodo e' corrotto*/

			/* il padre di <sostituto> e' lo stesso di <joinpoint>, poiche' il primo sostituira' il secondo nell'albero */
			sostituto->parent=joinpoint->parent;
			/* sostituzione effettiva di <sostituto> a <joinpoint> tra i figli del padre di <joinpoint>*/
			padre=joinpoint->parent;
			for(i=0;i<padre->nfigli && (padre->sons)[i]!=joinpoint;i++);
			/* al termine del ciclo, (padre->sons)[i] contiene il puntatore a <joinpoint> che rappresenta il legame di parentela tra i due nodi: per recidere questo legame e sostituire effettivamente <sostituto> a <joinpoint> occorre semplicemente far puntare tale puntatore a <sostituto>*/
			(padre->sons)[i]=sostituto;
			/* <joinpoint> diventa dunque il primo figlio di <sostituto>*/
			(sostituto->sons)[0]=joinpoint;
			/* quindi <sostituto> ora e' il nuovo padre di <joinpoint>*/
			joinpoint->parent=sostituto;
			/* i figli di sostituto sono ovviamente 1, almeno per ora*/
			sostituto->nfigli=1;
			/* il padre del nuovo nodo di cui il chiamante ha chiesto l'inserimento e' dunque <sostituto>*/
			padre=sostituto;
	       }
	}

	/* alloca il nuovo figlio e controlla che l'allocazione vada a buon fine*/
	if((nuovo=malloc(sizeof(alberolkh)))==NULL)
	{
		return FALSE;
	}

	/* inizializza opportunamente i campi del nuovo figlio, se ci sono errori di allocazione dealloca tutto cio' che e' stato precedentemente allocato ed esce restituendo al chiamante FALSE */
	if((nuovo->mysubset=malloc(sizeof(Sij)))==NULL)
	{
		free(nuovo);
		return FALSE;
	}
	if((nuovo->sons=calloc(maxfigli,sizeof(alberolkh *)))==NULL)
	{
		free(nuovo->mysubset);
		free(nuovo);
		return FALSE;
	}
	nuovo->nfigli=0; /* il nuovo nodo e' ovviamente una foglia */
	/*il subset viene inizializzato con il valore specificato dal chiamante*/
	*(nuovo->mysubset)=insieme;
	nuovo->iscorrupted=TRUE; /* il path cui appartiene il nodo e' corrotto poiche' e' appena avvenuto un inserimento */

	/* collega il nodo al genitore */
	nuovo->parent=padre;
	/*collega il genitore al figlio, ponendo il primo puntatore libero della lista dei figli del padre al nuovo nodo appena creato (ovviamente tale puntatore e' (radice->sons)[radice->nfigli], poiche' nfigli contiene il numero attuale di figli del padre)*/
	(padre->sons)[padre->nfigli]=nuovo;
	/*incrementa il numero di figli del padre */
	(padre->nfigli)++;
	/*Corrompe fisicamente tutto il path da <nuovo> fino alla radice*/
	Corrompi(nuovo);
	return TRUE;
}


/* Scendi(): fa scendere di un livello il nodo <nodo>, creando un figlio identico a tale nodo come suo discendente */
int Scendi(alberolkh *nodo, int maxfigli)
{
	alberolkh *padre, *sostituto;
	int i;

	padre=nodo->parent;



	/* alloca il nodo che sostituira' <padre> e controlla che l'allocazione vada a buon fine*/
	if((sostituto=malloc(sizeof(alberolkh)))==NULL)
	{
		return FALSE;
	}
	/* alloca i campi del nuovo nodo, se ci sono errori di allocazione 	dealloca tutto cio' che e' stato precedentemente allocato ed esce restituendo al chiamante FALSE */
	if((sostituto->sons=calloc(maxfigli,sizeof(alberolkh *)))==NULL)
	{
		free(sostituto);
		return FALSE;
	}

 

	/* il subset di <figlio> e' lo stesso di <padre>, quello di <padre> non esiste, essendo adesso <padre> un nodo interno*/
	sostituto->mysubset=NULL;
	nodo->iscorrupted=TRUE; /*ovviamente il path a cui appartiene il nodo e' corrotto*/
	/* il padre di <figlio> e' ovviamente <padre> */
	nodo->parent=sostituto;
	sostituto->parent=padre;
	/* e il primo figlio di <padre> e' ovviamente <figlio> */
	(sostituto->sons)[0]=nodo;
 	/* i figli di <padre> sono ovviamente 1, almeno per ora*/
	sostituto->nfigli=1;
	/* i figli di <figlio> invece sono 0, essendo questo una foglia */
	nodo->nfigli=0;

	for(i=0;i<padre->nfigli && (padre->sons)[i]!=nodo;i++);
	/* al termine del ciclo, (padre->sons)[i] contiene il puntatore a <joinpoint> che rappresenta il legame di parentela tra i due nodi: per recidere questo legame e sostituire effettivamente <sostituto> a <joinpoint> occorre semplicemente far puntare tale puntatore a <sostituto>*/
	(padre->sons)[i]=sostituto;


	/* corrompe il path di padre */
	Corrompi(sostituto);
	/* ritorna TRUE poiche' e' andato tutto per il verso giusto */
	return TRUE;
}


/* Sali(): fa salire di un livello il nodo <nodo>, sostituendolo al padre */
void Sali(alberolkh *nodo)
{
	alberolkh *padre, *figlio, *nonno;
	int i;

	figlio=nodo;
	padre=nodo->parent;

	/* collega <figlio> al padre di <padre>*/
	figlio->parent=padre->parent;


	nonno=padre->parent;
	/* al termine del ciclo, (nonno->sons)[i] conterra' il puntatore a <padre> che rappresenta il legame di parentela tra i due nodi: per recidere questo legame e sostituire effettivamente <figlio> a <padre> occorre semplicemente far puntare tale puntatore a <figlio>*/
	for(i=0;i<nonno->nfigli && (nonno->sons)[i]!=padre;i++);

	/* lega il padre del padre al nodo <figlio> */
	(nonno->sons)[i]=figlio;

	/* corrompe il path di <figlio>*/
	//figlio->iscorrupted=TRUE;
	Corrompi(nonno);

	/* adesso <padre> e' privo di legami e di utilita', e puo' essere cancellato */

	free(padre->sons);
	free(padre);
}


long Profonditadi(alberolkh *radice)
{
	long i;
	
	i=0;
	while(radice!=NULL)
	{
		radice=radice->parent;
		i++;
	}
	return i;
}


int SingleBalance(alberolkh *radice, int maxfigli)
{
	alberolkh *massimo,*minimo, *max1, *max2;
	int livello1,livello2, fratelli;
	int i,j;
	//int flag=0;

	fratelli=0;
	livello1=livello2=0;

	/* trova la foglia piu' alta e piu' bassa (<massimo> e', fra tutte le foglie di altezza massima, quella con antenato corrotto piu' prossimo ed eventualmente con piu' fratelli di tutte le altre, <minimo> il nodo di altezza minore in cui e' possibile inserire) */
	minimo=LastLeaf(radice,&livello1,maxfigli);
	//massimo=MaxLeaf(radice,&livello1,&fratelli);

	/* Sceglie massimo all'interno dell'intero albero*/
	max1=MaxLeaf(radice,&livello1,&fratelli);
	fratelli=0;
	/* cerca un nodo della medesima altezza del massimo dell'albero all'interno del path in cui si vuole inserire per bilanciare (poiche' per tale path sarebbe comunque necessario il rekeying,
	se esiste una foglia di altezza pari a quella della massima foglia dell'albero, allora conviene che tale foglia e i suoi fratelli siano utilizzati nell'operazione di bilanciamento anziche' corrompere un nuovo path */
	max2=MaxLeaf(minimo,&livello2,&fratelli);
	//printf("max1=%ld, max2=%ld\n",livello1,livello2);
	//printf("max1=%ld, max2=%ld alt max1=%ld altmax2=%ld, Deep(minimo)=%ld confronto=%ld\n",livello1,livello2,Quantoealto(max1),Quantoealto(max2),Deep(radice)-Deep(minimo),(livello2+Deep(minimo)));
	
	/* sulla base dei parametri esposti nel precedente commento, sceglie quale delle due foglie (e dei relativi fratelli) saranno utilizzati nel bilanciamento.  */
	if((max2!=minimo) && (max2->parent!=minimo) &&  livello1==Profonditadi(max2)) 
	{
		//printf("scelgo max2\n");
		massimo=max2;
	}
	else
	{
		//if((max2->parent==minimo))
		//printf(" scelgo perche' minimo\n");
		//printf("scelgo max1\n");
		massimo=max1;
	}
	
	

	

	/* fa puntare <massimo> al padre della foglia di altezza maggiore */
	massimo=massimo->parent;

	/* se il nodo di altezza minore in cui e' possibile inserire e' una foglia...*/
	if (IsLeaf(minimo))
	{
		/* fa scendere <minimo> di un livello */

		if(Scendi(minimo,maxfigli)==FALSE) /* se l'operazione non riesce...*/
		{
			return FALSE; /* ... restituisce FALSE per segnalare al chiamante il problema */
		}


		/* inserisce affianco di minimo, quindi come figlio del padre di minimo, che e' appena sceso di un livello */
		minimo=minimo->parent;
	}
	

	/* finche' al nodo <minimo> non e' piu' possibile aggiungere figli*/
	for(i=0,j=minimo->nfigli;i<(massimo->nfigli-1) && j<maxfigli;i++,j++)
	{
		/* poni l'i-esimo figlio di <massimo> come j-esimo figlio di <minimo> */
		(minimo->sons)[j]=(massimo->sons)[i];
		((minimo->sons)[j])->parent=minimo;
		((minimo->sons)[j])->iscorrupted=TRUE;
	}

	/* aggiorna il numero di figli di <minimo> */
	minimo->nfigli=j;

	/* se dopo lo scambio di figli a <massimo> e' rimasto un solo figlio...*/
	if(i==(massimo->nfigli-1))
	{
		/* ... fa salire quel figlio di un livello */
		Sali((massimo->sons)[i]);
	}
	/* altrimenti, se c'e' piu' di un figlio...*/
	else
	{
		/*... pone i figli rimasti di <massimo> a partire da (massimo->sons)[0] */
		for(j=0;i<massimo->nfigli;j++,i++)
		{
			(massimo->sons)[j]=(massimo->sons)[i];
			((massimo->sons)[j])->iscorrupted=TRUE;
		}
		/* aggiorna il numero di figli di <massimo> */
		massimo->nfigli=j;
	}

	/* corrompe i due path coinvolti nel bilanciamento */
	Corrompi(massimo);
	Corrompi(minimo);

	/* l'operazione e' andata a buon fine, torna TRUE per segnalarlo */
	return TRUE;

}

int Balance(alberolkh *radice, int sbilanciamento,int maxfigli)
{
	int eseguitabalance=1;
	int quantevolte = 0;

	while(!IsBalanced(radice, sbilanciamento, maxfigli))
	{
		quantevolte++;
		//printf ("\nBilanciato %d volte\n", quantevolte);
		eseguitabalance=TRUE;
		//printf("Bilancio!\n");
		if(SingleBalance(radice, maxfigli)==FALSE)
		{

			return FALSE;
		}
		//printf("Ho Bilanciato!\n");

	}

	

	return eseguitabalance;
}

void Stampa(alberolkh *testa)
{
	int i;



	if(testa->mysubset==NULL)
	{

	printf("Nodo interno Chiave %d\n",(int) testa->key[0]);
	}
	else
	{
	printf("Nodo foglia: i=%ld j=%ld\n",testa->mysubset->i,testa->mysubset->j);
	}

	if(testa->nfigli==0 && testa->mysubset==NULL)
	{
			printf("Figlio di %c: non c'e'!\n",testa->key[0]);
	}
	/*else if(testa->nfigli==0 && testa->mysubset!=NULL)
	{
		printf("Figlio di sij i=%ld j=%ld: non c'e'!\n",testa->mysubset->i,testa->mysubset->j);
	}*/


	for(i=0;i<testa->nfigli;i++)
	{
		printf("Figli di %c:\n" ,testa->key[0]);
		Stampa((testa->sons)[i]);

	}
}


/* Restitusce 1 se l'albero puntato da nodo ?? padre di sole foglie */
int isSemiLeaf(alberolkh *nodo)
{
	int i= 0;

	if (nodo == NULL) return 0;
	if (IsLeaf(nodo->sons[i])) return 0;

	for (i = 0; i < nodo->nfigli; i++)
	{
		if (!IsLeaf(nodo->sons[i]))
		{
			return 0;
		}
	}

	return 1;
}


/* Leafs(): ritorna una lista di Sij contenuto da tutte le foglie dell'albero radicato in <nodo> */
listasij *Leafs(alberolkh *nodo)
{
	listasij *subset, *subsetson;
	int i;

	if(nodo==NULL)
	{
		return NULL;
	}
	else if(IsLeaf(nodo) && nodo->mysubset!=NULL)
	{
		if((subset=malloc(sizeof(listasij)))==NULL)
		{
			return NULL;
		}
		subset->i=(nodo->mysubset)->i;
		subset->j=(nodo->mysubset)->j;
		subset->next=NULL;
		return subset;
	}
	else
	{
		subset=NULL;
		for(i=0;i<nodo->nfigli;i++)
		{
			subsetson=Leafs((nodo->sons)[i]);
			if(subsetson==NULL)
			{
				return NULL;
			}
			ListCat(&subset,subsetson);
		}
	}

	return subset;

}


/*NumFoglie(): conta il numero di foglie (intese come foglie contenenti subset) nel sottoalbero radicato in <radice>*/
long NumFoglie(alberolkh *radice)
{
	int i;
	long totale;

	totale=0;

	/* se la radice non esiste...*/
	if(radice == NULL)
	{
		/* ... allora non ci sono foglie nel sottoalbero */
		return 0;
	}
	/* se la radice e' una foglia (con subset)...*/
	if(IsLeaf(radice))
	{
		if( (radice->mysubset)!=NULL )
		{
			/*...allora il numero totale di foglie e' pari ad uno */
			return 1;
		}
		/* se e' una foglia senza subset, allora e' un nodo interno divenuto foglia, e dunque non va annoverato nel numero totale */
		else
		{
			return 0;
		}
	}

	/* altrimenti, per ogni figlio di <radice>...*/
	for(i=0;i<radice->nfigli;i++)
	{
		/*...calcola il numero di foglie contenute nel sottoalbero radicato nell'i-esimo figlio */
		totale+=NumFoglie((radice->sons)[i]);
	}

	return totale;
}

/* SimulaRekeying(): simula l'esecuzione dell'operazione di Rekeying secondo l'algoritmo  LKH, senza pero' procedere alla ridistribuzione delle chiavi */
long SimulaRekeying(alberolkh *radice, nodosd* utenti, long nutenti, long utenteiniziale)
{
	long som , i ,retVal;
	

	/* se la radice non esiste, oppure non e' corrotta...*/
	if(radice == NULL || !(radice->iscorrupted))
	{
		/* ... allora non occorre inviare messaggi, dunque il numero di mex e' 0*/
		return 0;
	}
	/* se la radice e' una foglia...*/
	if(IsLeaf(radice))
	{
		/*...allora, poiche' il genitore sta per subire il rekeying, la foglia non va piu' considerata corrotta */
		radice->iscorrupted=FALSE;
		return 0;
	}


	//printf("\nquesto sottoalbero ha %ld foglie\n",NumFoglie(radice));

	i=0;
	som=0;

	/* altrimenti, poiche' la radice ha vari figli, esegue il Rekeying su ognuno di essi */
	while(i < radice->nfigli)
	{
		retVal = SimulaRekeying((radice->sons)[i], utenti, nutenti,utenteiniziale);
		/* se si e' verificato qualche errore...*/
		if (retVal == -1)
		{
			/*...ritorna il codice di errore */
			return -1;
		}

		/* aggiunge il numero di messaggi necessari ad eseguire il rekeying sull'albero radicato in ((radice->sons)[i]) al numero totale di messaggi inviati*/

		som+=retVal;

		i++;
	}


	/* restituisce il numero sin qui trovato di messaggi piu' tanti messaggi quanti sono i  figli (la chiave attuale viene inviata crittata con le chiavi di ciascun nodo figlio)*/
	return som+radice->nfigli;
}


/* Rekeying(): Esegue l'operazione di Rekeying secondo l'algoritmo  LKH e restituisce il numero di messaggi utilizzati  nell'operazione */
long Rekeying(alberolkh *radice, nodosd* utenti, long nutenti, long utenteiniziale)
{
	long som , i ,retVal;
	listasij *destinatari,*d;
	chiave vecchia;


	/* se la radice non esiste, oppure non e' corrotta...*/
	if(radice == NULL || !(radice->iscorrupted))
	{
		/* ... allora non occorre inviare messaggi, dunque il numero di mex e' 0*/
		return 0;
	}
	/* se la radice e' una foglia...*/
	if(IsLeaf(radice))
	{
		/*...allora, poiche' il genitore sta per subire il rekeying, la foglia non va piu' considerata corrotta */
		radice->iscorrupted=FALSE;
		return 0;
	}


	//printf("\nquesto sottoalbero ha %ld foglie\n",NumFoglie(radice));

	i=0;
	som=0;

	/* altrimenti, poiche' la radice ha vari figli, esegue il Rekeying su ognuno di essi */
	while(i < radice->nfigli)
	{
		retVal = Rekeying((radice->sons)[i], utenti, nutenti,utenteiniziale);
		/* se si e' verificato qualche errore...*/
		if (retVal == -1)
		{
			/*...ritorna il codice di errore */
			return -1;
		}

		/* aggiunge il numero di messaggi necessari ad eseguire il rekeying sull'albero radicato in ((radice->sons)[i]) al numero totale di messaggi inviati*/

		som+=retVal;

		i++;
	}



	/* salva in <vecchia> la vecchia chiave di <radice> */
	memcpy(vecchia,radice->key,SHA_DIGEST_LENGTH);
	/* poiche' questa chiave verra' sostituita dalla nuova, elimina la vecchia chiave dagli utenti che la posseggono ancora */
	EliminaChiave(vecchia,utenti,nutenti,utenteiniziale);
	/* calcola la nuova chiave di <radice>*/
	if ( !GetNewKey(radice->key) )
	{
		return -1;
	}


	/* seleziona gli Sij destinatari della nuova chiave (ovvero tutti gli Sij contenuti come foglie nell'albero radicato in <radice>) */
	destinatari=Leafs(radice);




	/*  simula l'invio in broadcast della nuova chiave di <radice> crittata con le chiavi di ognuno dei figli di <radice> stessa */
	if(Broadcast(radice->key,destinatari,utenti,nutenti)==-1)
	{
		return -1;
	}


	/* essendo stata assegnata una nuova chiave a <radice>, tale nodo non e' piu' da considerarsi corrotto */
	radice->iscorrupted=FALSE;

	/* dealloca i destinatari, che ormai non servono piu' */

	while(destinatari!=NULL)
	{
		d=destinatari;
		destinatari=destinatari->next;
		free(d);
	}


	/* restituisce il numero sin qui trovato di messaggi piu' uno (il messaggio appena inviato in bcast) */
	//return som+1;

	/* restituisce il numero sin qui trovato di messaggi piu' tanti messaggi quanti sono i  figli (la chiave attuale viene inviata crittata con le chiavi di ciascun nodo figlio)*/
	return som+radice->nfigli;
}
